@@ -1,7 +1,11 @@
 package Apache::Registry;
 use Apache ();
 #use strict; #eval'd scripts will inherit hints
-use Apache::Constants qw(:common &OPT_EXECCGI &REDIRECT);
+use Apache::Constants qw(:common &OPT_EXECCGI);
+
+BEGIN {
+    OPT_EXECCGI() if $ENV{MOD_PERL}; #preload, :common are alread pre-loaded
+}
 
 $Apache::Registry::VERSION = '2.01';
 
@@ -54,13 +58,15 @@ sub handler {
 	my $mtime = -M _;
 
 	my $uri = $r->uri;
-	$uri = "/__INDEX__" if $uri eq "/";
 	# turn into a package name
 	$r->log_error(sprintf "Apache::Registry::handler examining %s",
 		      $uri) if $Debug && $Debug & 4;
-	my $script_name = $r->path_info ?
-	    substr($uri, 0, length($uri)-length($r->path_info)) :
-		$uri;
+	my $path_info = $r->path_info;
+	my $script_name = $path_info && $uri =~ /$path_info$/ ?
+	    substr($uri, 0, length($uri)-length($path_info)) :
+	    $uri;
+
+	$script_name =~ s:/+$:/__INDEX__:;
 
 	if($Apache::Registry::NameWithVirtualHost) {
 	    my $name = $r->get_server_name;
@@ -68,14 +74,14 @@ sub handler {
 	}
 
 	# Escape everything into valid perl identifiers
-	$script_name =~ s/([^A-Za-z0-9\/])/sprintf("_%2x",unpack("C",$1))/eg;
+	$script_name =~ s/([^A-Za-z0-9_\/])/sprintf("_%2x",unpack("C",$1))/eg;
 
 	# second pass cares for slashes and words starting with a digit
 	$script_name =~ s{
 			  (/+)       # directory
 			  (\d?)      # package's first character
 			 }[
-			   "::" . ($2 ? sprintf("_%2x",unpack("C",$2)) : "")
+			   "::" . (length $2 ? sprintf("_%2x",unpack("C",$2)) : "")
 			  ]egx;
 
 	my $package = "Apache::ROOT$script_name";
@@ -131,6 +137,8 @@ sub handler {
 	    $Apache::Registry->{$package}{'mtime'} = $mtime;
 	}
 
+	my $old_status = $r->status;
+
 	my $cv = \&{"$package\::handler"};
 	eval { &{$cv}($r, @_) } if $r->seqno;
 	$r->chdir_file("$Apache::Server::CWD/");
@@ -155,7 +163,7 @@ sub handler {
 #		return REDIRECT;
 #	    }
 #	}
-	return $r->status;
+	return $r->status($old_status);
     } else {
 	return NOT_FOUND unless $Debug && $Debug & 2;
 	return Apache::Debug::dump($r, NOT_FOUND);
@@ -220,11 +228,10 @@ Apache::Registry - Run unaltered CGI scrips under mod_perl
  PerlModule Apache::Registry
 
  <Location /perl>
- SetHandler perl-script
- PerlHandler Apache::Registry
- Options ExecCGI 
- ...
- </Directory>
+   SetHandler perl-script
+   PerlHandler Apache::Registry
+   Options ExecCGI 
+ </Location>
 
 =head1 DESCRIPTION
 
