@@ -1,7 +1,7 @@
 package Apache::src;
 
 use strict;
-use vars qw($VERSION $AUTOLOAD);
+use vars qw($VERSION);
 use File::Path ();
 use IO::File ();
 use Cwd ();
@@ -15,6 +15,31 @@ use Config;
 $VERSION = '0.01';
 sub IS_MOD_PERL_BUILD () {-e "../lib/mod_perl.pm"}
 my $Is_Win32 = ($^O eq "MSWin32");
+$Apache::src::APXS ||= "";
+
+sub apxs {
+    my $self = shift;
+    eval { require Apache::MyConfig };
+    my $apxs;
+    for ($Apache::src::APXS,
+	 $Apache::MyConfig::Setup{'APXS'},
+	 `which apxs`,
+	 "/usr/local/apache/bin/apxs")
+      {
+	  next unless ($apxs = $_);
+	  chomp $apxs;
+	  last if -x $apxs;
+      }
+    return "" unless $apxs and -x $apxs;
+    `$apxs @_`;
+}
+
+sub apxs_cflags {
+    my $cflags = __PACKAGE__->apxs(-q => 'CFLAGS');
+    #$cflags =~ s/-D\w+=\".*\"//g; #get rid of -Ds with quotes
+    $cflags =~ s/\"/\\\"/g;
+    $cflags;
+}
 
 sub new {
     my $class = shift;
@@ -124,7 +149,7 @@ sub asrc {
     return $d if -e "$d/httpd.h";
     return "$d/include" if -e "$d/include/httpd.h";
     return "$d/main" if -e "$d/main/httpd.h";
-    return undef;
+    return Apache::src->apxs(-q => 'INCLUDEDIR');
 }
 
 sub module_magic_number {
@@ -167,6 +192,8 @@ sub httpd_version {
     my($string, $extra, @vers);
 
     while(<$fh>) {
+	next unless /^#define/;
+	s/SERVER_PRODUCT \"/\"Apache/; #1.3.13+
 	next unless s/^#define\s+SERVER_(BASE|)VERSION\s+"(.*)\s*".*/$2/;
 	chomp($string = $_);
 
@@ -190,16 +217,37 @@ sub httpd_version {
     return $version;
 }
 
+sub find_in_inc {
+    my $name = shift;
+    for (@INC) {
+	my $file;
+	if (-e ($file = "$_/auto/Apache/$name")) {
+	    return $file;
+	}
+    }
+}
+
+sub otherldflags {
+    my $self = shift;
+    my @ldflags = ();
+
+    if ($^O eq "aix") {
+	if (my $file = find_in_inc("mod_perl.exp")) {
+	    push @ldflags, "-bI:" . $file;
+	}
+	my $httpdexp = $self->apxs(-q => 'LIBEXECDIR') . "/httpd.exp";
+	push @ldflags, "-bI:$httpdexp" if -e $httpdexp;
+    }
+    return join(' ', @ldflags);
+}
+
 sub typemaps {
     my $typemaps = [];
     
-    for (@INC) {
-	my $file;
-	if (-e ($file = "$_/auto/Apache/typemap")) {
-	    push @$typemaps, $file;
-	    last;
-	}
+    if (my $file = find_in_inc("typemap")) {
+	push @$typemaps, $file;
     }
+
     if(IS_MOD_PERL_BUILD) {
 	push @$typemaps, "../Apache/typemap";
     }
@@ -212,8 +260,8 @@ sub inc {
     my $src  = $self->dir;
     my $main = $self->main;
     my $os = $Is_Win32 ? "win32" : "unix";
-    my @inc = ("-I$src", "-I$src/modules/perl", "-I$main");
-    for ("src/regex", "$src/os/$os") {
+    my @inc = (); 
+    for ($src, "$src/modules/perl", $main, "$src/regex", "$src/os/$os") {
 	push @inc, "-I$_" if -d $_;
     }
     my $ssl_dir = "$src/../ssl/include";
@@ -222,9 +270,17 @@ sub inc {
 	$ssl_dir = "$Apache::MyConfig::Setup{SSL_BASE}/include";
     }
     push @inc, "-I$ssl_dir" if -d $ssl_dir;
+    my $ainc = $self->apxs(-q => 'INCLUDEDIR');
+    push @inc, "-I$ainc" if -d $ainc;
     return "@inc";
 }
 
+sub ccflags {
+    my $self = shift;
+    my $cflags = $Config{'ccflags'};
+    join " ", $cflags, $self->apxs(-q => 'CFLAGS');
+}
+
 sub define {
     my $self = shift;
     if($Config{usethreads}) {
@@ -274,7 +330,7 @@ source code.
 Create an object blessed into the B<Apache::src> class.
 
  my $src = Apache::src->new;
- 
+
 =item dir
 
 Top level directory where source files are located.
@@ -338,6 +394,28 @@ Example:
 
  my $v = $src->httpd_version;
 
+=item otherldflags
+
+Return other ld flags for MakeMaker's B<dynamic_lib> argument to
+C<WriteMakefile>. This might be needed on systems like AIX that need
+special flags to the linker to be able to reference mod_perl or httpd
+symbols.
+
+Example:
+
+ use ExtUtils::MakeMaker;
+
+ use Apache::src ();
+
+ WriteMakefile(
+     'NAME'        => 'Apache::Module',
+     'VERSION'     => '0.01', 
+     'INC'         => Apache::src->new->inc,	      
+     'dynamic_lib' => {
+	 'OTHERLDFLAGS' => Apache::src->new->otherldflags,
+     },
+ );
+
 =back
 
 
